#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/stat.h>
#include <stdarg.h>
#include <errno.h>
#include <dirent.h>
#include "listfiles.h"


/***********************************************************************
 *                                                                     *
 *                         Pointer Buffer Code                         *
 *                                                                     *
 ***********************************************************************/

typedef struct ptrbuf ptrbuf_t;

#define _PTRBUF_PROTO static

_PTRBUF_PROTO ptrbuf_t *ptrbuf_new(size_t blksize);
_PTRBUF_PROTO void ptrbuf_free(ptrbuf_t *B, void ***Rd, size_t *Rl);

_PTRBUF_PROTO void ptrbuf_collect(ptrbuf_t *B, void ***Rd, size_t *Rl);

     // always puts to the end of the buffer
_PTRBUF_PROTO void ptrbuf_put(ptrbuf_t *B, void *d);

     // gets from the start of the buffer
_PTRBUF_PROTO void* ptrbuf_get(ptrbuf_t *B);

     // gets from the end of the buffer
_PTRBUF_PROTO void* ptrbuf_pop(ptrbuf_t *B);

struct pbufnode { 
  struct pbufnode *next;
  size_t len, start;
  void* data[];
};

struct ptrbuf {
  struct pbufnode *first, *last;
  size_t blksize;
};


_PTRBUF_PROTO ptrbuf_t *ptrbuf_new(size_t blksize)
{
  ptrbuf_t *B;
  B= malloc(sizeof(*B));
  B->first= B->last= NULL;
  B->blksize= blksize;
  return B;
}


_PTRBUF_PROTO void ptrbuf_free(ptrbuf_t *B, void ***Rd, size_t *Rl)
{
  struct pbufnode *P,*N;
  if (Rd || Rl) ptrbuf_collect(B, Rd, Rl);
  for(P=B->first;P;P=N)
  {
    N= P->next;
    free(P);
  }
  free(B);
}

_PTRBUF_PROTO void ptrbuf_collect(ptrbuf_t *B, void ***Rd, size_t *Rl)
{
  size_t S;
  struct pbufnode *P;
  void **R;

  if (!Rd && !Rl) return ;

  S= 0;
  for(P=B->first;P;P=P->next) S+= P->len;

  if (S==0) { if (Rd) *Rd= NULL; if (Rl) *Rl= 0; return ; }
  if (!Rd) { if (Rl) *Rl= S; return ; }

  R= malloc(S*sizeof(void*));
  S= 0;
  for(P=B->first;P;P=P->next)
  {
    memcpy(R+S, P->data+P->start, P->len*sizeof(void*));
    S+= P->len;
  }

  *Rd= R;
  if (Rl) *Rl= S;
}

_PTRBUF_PROTO void ptrbuf_put(ptrbuf_t *B, void *d)
{
  struct pbufnode *P;

  P= B->last;
  if (!P || (P->len+P->start)==B->blksize)
  {
      P= malloc(sizeof(*P)+B->blksize*sizeof(void*));
      P->len= P->start= 0;
      P->next= NULL;
      if (!B->last) { B->last= B->first= P; }
      else { B->last->next= P; B->last= P; }
 }
 P->data[P->len+P->start]= d;
 P->len++;
}

_PTRBUF_PROTO void* ptrbuf_get(ptrbuf_t *B)
{
  struct pbufnode *P;
  void *R;

  P= B->first;
  if (!P) return NULL;
  R= P->data[P->start];
  P->start++;
  P->len--;
  if (P->len==0)
  {
    B->first= P->next;
    if (!B->first) B->last= NULL;
    free(P);
  }
  return R;
}

_PTRBUF_PROTO void* ptrbuf_pop(ptrbuf_t *B)
{
  struct pbufnode *P,**T,*Y;
  void *R;

  P= B->last;
  if (!P) return NULL;
  R= P->data[P->start+P->len-1];
  P->len--;
  if (P->len==0)
  {
    for(Y=B->first,T=&B->first;Y!=P;)
    {
      T= &Y->next;
      Y= Y->next;
    }
    *T= NULL;
    free(P);
    if (!B->first) B->last= NULL;
  }
  return R;
}


/***********************************************************************
 *                                                                     *
 *                            Main Library                             *
 *                                                                     *
 ***********************************************************************/

typedef struct {
  ptrbuf_t *dirbuf;
  ptrbuf_t *filbuf;
  file_ent_t fe;
  char *fn;
} LF_t;

static void (*listfilesError)(void*,const char*,va_list);
static void *listfilesErrorArg;
static int listfilesErrorInit;

static file_ent_t *lfDestroy(LF_t *LF);
static int lfStat(LF_t *LF, char *fn);
static file_ent_t *lfDestroy(LF_t *LF);

void listfilesSetError(void (*f)(void*,const char *,va_list), void *a)
{
  listfilesError= f;
  listfilesErrorArg= a;
  listfilesErrorInit= 1;
}

static void listfilesErrorStd(void *dummy, const char *fmt, va_list ap)
{
  vfprintf(stderr,fmt, ap);
}

static void lfError(const char *fmt, ...)
{
  va_list ap;
  if (!listfilesErrorInit)
  {
    listfilesErrorInit= 1;
    listfilesError= listfilesErrorStd;
    listfilesErrorArg= NULL;
  }
  if (!listfilesError) return ;
  va_start(ap,fmt);
  (*listfilesError)(listfilesErrorArg, fmt, ap);
  va_end(ap);
}

static int lfCheckDir(LF_t *LF, int force)
{
  if (S_ISDIR(LF->fe.sbuf.st_mode)) return 0;
  if (force)
    lfError("%s isn't a directory\n", LF->fn);
  return 1;
}

static char *mkpath(const char *x1, const char *x2)
{
  char *R;
  R= malloc(strlen(x1)+1+strlen(x2)+1);
  sprintf(R,"%s/%s",x1,x2);
  return R;
}

static void lfGetFiles(LF_t *LF, const char* dirname)
{
  DIR *dir;
  struct dirent *de;
  dir= opendir(dirname);
  if (!dir)
  {
    lfError("opendir(%s): %s\n", dirname, strerror(errno));
    return ;
  }
  while((de=readdir(dir)))
  {
    if (!strcmp(de->d_name,".") || !strcmp(de->d_name,"..")) continue;
    ptrbuf_put(LF->filbuf, mkpath(dirname, de->d_name));
  }
  closedir(dir);
}

static file_ent_t *lfGetEnt(LF_t *LF)
{
  mode_t m; int t;
  char *fn,*dn;
again:
  fn= ptrbuf_get(LF->filbuf);
  if (!fn)
  {
    dn= ptrbuf_get(LF->dirbuf);
    if (!dn) return lfDestroy(LF);
    lfGetFiles(LF,dn);
    free(dn);
    goto again;
  }
  if (lfStat(LF, fn)) { free(fn); goto again; }
  if (!lfCheckDir(LF,0))
  {
    ptrbuf_put(LF->dirbuf, strdup(fn));
  }
  m= LF->fe.sbuf.st_mode;
  if (S_ISDIR(m)) t= 'D';
  else if (S_ISREG(m)) t= 'F';
  else if (S_ISCHR(m)) t= 'C';
  else if (S_ISBLK(m)) t= 'B';
  else if (S_ISFIFO(m)) t= 'P';
  else if (S_ISLNK(m)) t= 'L';
  else if (S_ISSOCK(m)) t= 'S';
  else t= 'U';
  LF->fe.type= t;
  LF->fe.path= LF->fn;
  LF->fe.priv= LF;
  return &LF->fe;
}

file_ent_t *listfilesBegin(const char *_dirname)
{
  LF_t *LF;
  char *dirname;
  if (!_dirname || !_dirname[0]) return NULL;
  LF= calloc(1,sizeof(*LF));
  LF->dirbuf= ptrbuf_new(50);
  LF->filbuf= ptrbuf_new(50);

  dirname= strdup(_dirname);
  if (lfStat(LF, dirname)) return lfDestroy(LF); 
  if (lfCheckDir(LF, 1))   return lfDestroy(LF);
  LF->fn= NULL;
  ptrbuf_put(LF->filbuf, dirname);
  // if (!strcmp(dirname,".")) ptrbuf_put(LF->dirbuf, dirname);
  // else ptrbuf_put(LF->filbuf, dirname);
  return lfGetEnt(LF);
}

static int lfStat(LF_t *LF, char *fn)
{
  if (LF->fn) free(LF->fn);
  LF->fn= fn;
  if (!lstat(fn, &LF->fe.sbuf)) return 0;
  lfError("stat(%s): %s\n", fn, strerror(errno));
  return 1;
}



file_ent_t *listfilesNext(file_ent_t *fe)
{
  if (!fe) return NULL;
  return lfGetEnt(fe->priv);
}

file_ent_t *listfilesEnd(file_ent_t *fe)
{
  return lfDestroy(fe->priv);
}

static file_ent_t *lfDestroy(LF_t *LF)
{
  char *p;
  while((p=ptrbuf_pop(LF->dirbuf))) free(p);
  while((p=ptrbuf_pop(LF->filbuf))) free(p);
  ptrbuf_free(LF->dirbuf, NULL, NULL);
  ptrbuf_free(LF->filbuf, NULL, NULL);
  memset(LF, 0, sizeof(*LF));
  free(LF);
  return NULL;
}

